@@ -1,9 +1,11 @@
 from . import v0
 from . import v1
 from . import v2
+from . import v3
 
 SCHEMAS = [
     v0.V0,
     v1.V1,
     v2.V2,
+    v3.V3,
 ]
